#define _GNU_SOURCE
#include<stdio.h>
#include<string.h>
#include<dirent.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<zip.h>
#include<libtar.h>
#include<bzlib.h>
#include<zlib.h>

/*
walks through the directory and applies a function with one
parameter "something" to each file (filename),
can be useful for other compression methods in the future, perhaps
*/

static void traverse_directory(const char * directory, void * something, void (*func)(void *,const char *) )
{
    DIR * dp;
    struct dirent * dirp;
    char complete_name[BUFSIZ];
    char * end;

    dp = opendir(directory);
    while(dirp = readdir(dp))
        if(dirp->d_type == DT_REG)
	{
	    complete_name[0] = '\0';
            end = stpcpy(complete_name, directory);
	    if( end[-1] != '/' )
	    {
	        end = stpcpy(end, "/");
	    }
	    end = stpcpy(end, dirp->d_name);

	    func(something, complete_name);
	}    
    closedir(dp);
}

void add_to_zip(void * z, const char * filename)
{
    struct zip_source *s;

    s = zip_source_file( (struct zip *)z, filename, 0, 0);
    zip_add( (struct zip *)z, filename, s);
}


int create_zip(const char * archive_name, const char * directory)
{
    struct zip * z;

    z = zip_open(archive_name, ZIP_CREATE, NULL);
    traverse_directory(directory, z, add_to_zip);
    zip_close(z);

    return 0;
}

int create_tar(const char * archive_name, const char * directory)
{
    TAR *tar; 

    tar_open(&tar, (char *)archive_name, NULL, O_WRONLY | O_CREAT, 0644, TAR_GNU);
    tar_append_tree(tar, (char *)directory, ".");
    tar_close(tar);
    return 0;
}

int create_targz(const char * archive_name, const char * directory)
{
    char name_wo_gz[BUFSIZ];
    char buf[BUFSIZ];
    FILE * f;
    ssize_t bytesRead;
    gzFile gz;

    strcpy(name_wo_gz, archive_name);
    strrchr(name_wo_gz,'.')[0] = '\0';
    create_tar(name_wo_gz, directory);

    f = fopen(name_wo_gz, "r");
    gz = gzopen(archive_name, "w");

    while( (bytesRead = fread(buf, 1, BUFSIZ, f)) > 0 )
    {
        gzwrite(gz, buf, bytesRead);
    }
    gzclose(gz);
    fclose(f);
    remove(name_wo_gz);

    return 0;
}

int create_tarbz2(const char * archive_name, const char * directory)
{
    char name_wo_bz2[BUFSIZ];
    char buf[BUFSIZ];
    FILE * f;
    ssize_t bytesRead;
    int tarFD;
    int bzError;
    BZFILE * bz;
#define BLOCK_MULTIPLIER 7    
    
    strcpy(name_wo_bz2, archive_name);
    strrchr(name_wo_bz2,'.')[0] = '\0';
    create_tar(name_wo_bz2, directory);

    tarFD = open(name_wo_bz2, O_RDONLY);
    f = fopen(archive_name, "w");
    bz = BZ2_bzWriteOpen(&bzError, f, BLOCK_MULTIPLIER, 0, 0);

    while ( (bytesRead = read(tarFD,buf,BUFSIZ)) > 0)
    {
        BZ2_bzWrite(&bzError, bz, buf, bytesRead);
    }
    BZ2_bzWriteClose(&bzError, bz, 0, NULL, NULL);
    
    close(tarFD);
    remove(name_wo_bz2);
    return 0;
}

int create_archive(const char * archive_name, const char * directory)
{
    if (strstr(archive_name, ".zip"))
        return create_zip(archive_name, directory);
	
    if (strstr(archive_name, ".tar.gz"))
        return create_targz(archive_name, directory);

    if (strstr(archive_name, ".tar.bz2"))
        return create_tarbz2(archive_name, directory);

    if (strstr(archive_name, ".tar"))
        return create_tar(archive_name, directory);

    fprintf(stderr,"error: unknown archive type\n");
    return 1;
}

int main(int argc, char ** argv)
{
    if (argc != 3)
    {
        fprintf(stderr,"usage: create_archive archivename directory\n");
        return 1;
    }

    return create_archive(argv[1], argv[2]);
    
}
